﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment.Operators {
  [Item("ApproximateLocalSearch", "The approximate local search is described in Mateus, G., Resende, M., and Silva, R. 2011. GRASP with path-relinking for the generalized quadratic assignment problem. Journal of Heuristics 17, Springer Netherlands, pp. 527-565.")]
  [StorableClass]
  public class ApproximateLocalSearch : SingleSuccessorOperator, IProblemInstanceAwareGQAPOperator,
    IQualityAwareGQAPOperator, IGQAPLocalImprovementOperator, IStochasticOperator {
    public IProblem Problem { get; set; }
    public Type ProblemType {
      get { return typeof(GQAP); }
    }

    public ILookupParameter<GQAPInstance> ProblemInstanceParameter {
      get { return (ILookupParameter<GQAPInstance>)Parameters["ProblemInstance"]; }
    }
    public ILookupParameter<IntegerVector> AssignmentParameter {
      get { return (ILookupParameter<IntegerVector>)Parameters["Assignment"]; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<Evaluation> EvaluationParameter {
      get { return (ILookupParameter<Evaluation>)Parameters["Evaluation"]; }
    }
    public IValueLookupParameter<IntValue> MaximumIterationsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumIterations"]; }
    }
    public ILookupParameter<IntValue> EvaluatedSolutionsParameter {
      get { return (ILookupParameter<IntValue>)Parameters["EvaluatedSolutions"]; }
    }
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }
    public IValueLookupParameter<IntValue> MaximumCandidateListSizeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumCandidateListSize"]; }
    }
    public IValueLookupParameter<PercentValue> OneMoveProbabilityParameter {
      get { return (IValueLookupParameter<PercentValue>)Parameters["OneMoveProbability"]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters["Results"]; }
    }

    [StorableConstructor]
    protected ApproximateLocalSearch(bool deserializing) : base(deserializing) { }
    protected ApproximateLocalSearch(ApproximateLocalSearch original, Cloner cloner) : base(original, cloner) { }
    public ApproximateLocalSearch()
      : base() {
      Parameters.Add(new LookupParameter<GQAPInstance>("ProblemInstance", GQAP.ProblemInstanceDescription));
      Parameters.Add(new LookupParameter<IntegerVector>("Assignment", GQAPSolutionCreator.AssignmentDescription));
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", ""));
      Parameters.Add(new LookupParameter<Evaluation>("Evaluation", GQAP.EvaluationDescription));
      Parameters.Add(new ValueLookupParameter<IntValue>("MaximumIterations", "The maximum number of iterations that should be performed."));
      Parameters.Add(new LookupParameter<IntValue>("EvaluatedSolutions", "The number of evaluated solution equivalents."));
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MaximumCandidateListSize", "The maximum number of candidates that should be found in each step.", new IntValue(10)));
      Parameters.Add(new ValueLookupParameter<PercentValue>("OneMoveProbability", "The probability for performing a 1-move, which is the opposite of performing a 2-move.", new PercentValue(.5)));
      Parameters.Add(new LookupParameter<ResultCollection>("Results", "The result collection that stores the results."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ApproximateLocalSearch(this, cloner);
    }

    /// <summary>
    /// The implementation differs slightly from Mateus et al. in that the maximumIterations parameter defines a cap
    /// on the number of steps that the local search can perform. While the maxSampleSize parameter corresponds to
    /// the maxItr parameter defined by Mateus et al.
    /// </summary>
    /// <param name="random">The random number generator to use.</param>
    /// <param name="assignment">The equipment-location assignment vector.</param>
    /// <param name="quality">The solution quality.</param>
    /// <param name="evaluation">The evaluation result of the solution.</param>
    /// <param name="maxCLS">The maximum number of candidates that should be found in each step.</param>
    /// <param name="maximumIterations">The maximum number of iterations that should be performed each time the candidate list is generated.</param>
    /// <param name="problemInstance">The problem instance that contains the data.</param>
    /// <param name="oneMoveProbability">The probability for performing a 1-move, which is the opposite of performing a 2-move.</param>
    public static void Apply(IRandom random, IntegerVector assignment,
      DoubleValue quality, ref Evaluation evaluation, IntValue maxCLS, IntValue maximumIterations,
      GQAPInstance problemInstance, PercentValue oneMoveProbability) {
      var capacities = problemInstance.Capacities;
      var demands = problemInstance.Demands;
      //var weights = problemInstance.Weights;
      //var distances = problemInstance.Distances;
      //var installationCosts = problemInstance.InstallationCosts;
      while (true) {
        int count = 0;
        var CLS = new List<Tuple<NMove, double, Evaluation>>();
        double sum = 0.0;
        do {
          NMove move;
          if (random.NextDouble() < oneMoveProbability.Value)
            move = StochasticNMoveSingleMoveGenerator.GenerateExactlyN(random, assignment, 1, capacities);
          else move = StochasticNMoveSingleMoveGenerator.GenerateExactlyN(random, assignment, 2, capacities);
          
          var moveEval = GQAPNMoveEvaluator.Evaluate(move, assignment, evaluation, problemInstance);
          double moveQuality = problemInstance.ToSingleObjective(moveEval);

          if (moveEval.ExcessDemand <= 0.0 && moveQuality < 0.0) {
            CLS.Add(Tuple.Create(move, moveQuality, moveEval));
            sum += 1.0 / moveQuality;
          }
          count++;
        } while (CLS.Count < maxCLS.Value && count < maximumIterations.Value);

        if (CLS.Count == 0)
          return; // END
        else {
          var ball = random.NextDouble() * sum;
          var selected = CLS.Last();
          foreach (var candidate in CLS) {
            ball -= 1.0 / candidate.Item2;
            if (ball <= 0.0) {
              selected = candidate;
              break;
            }
          }
          NMoveMaker.Apply(assignment, selected.Item1);
          quality.Value += selected.Item2;
          evaluation = selected.Item3;
        }
      }
    }

    public override IOperation Apply() {
      var evaluation = EvaluationParameter.ActualValue;
      Apply(RandomParameter.ActualValue,
        AssignmentParameter.ActualValue,
        QualityParameter.ActualValue,
        ref evaluation,
        MaximumCandidateListSizeParameter.ActualValue,
        MaximumIterationsParameter.ActualValue,
        ProblemInstanceParameter.ActualValue,
        OneMoveProbabilityParameter.ActualValue);
      EvaluationParameter.ActualValue = evaluation;
      return base.Apply();
    }
  }
}
